# test_prompt_processors.py

import pytest
from dotenv import load_dotenv
import sys
import os

sys.path.append(os.path.join(os.path.dirname(__file__), '..', 'Modules'))

from PromptProcessorFactory import PromptProcessorFactory

load_dotenv()

def test_create_claude_prompt_processor():
    api_key = os.getenv("ANTHROPIC_API_KEY")
    processor = PromptProcessorFactory.create_prompt_processor("claude-3-haiku-20240307", api_key=api_key)
    assert processor.api_key == api_key
    assert processor.model == "claude-3-haiku-20240307"

    response = processor.generate_response("Hello, how are you?")
    assert isinstance(response, str)

def test_create_gpt_prompt_processor():
    api_key = os.getenv("OPENAI_API_KEY")
    processor = PromptProcessorFactory.create_prompt_processor("gpt-4o", api_key=api_key)
    assert processor.api_key == api_key

    response = processor.generate_response("Hello, how are you?", model="gpt-4o")
    assert isinstance(response, str)

def test_create_gemini_prompt_processor():
    api_key = os.getenv("GOOGLE_API_KEY")
    processor = PromptProcessorFactory.create_prompt_processor("models/gemini-1.5-flash-latest", api_key=api_key)
    assert processor.api_key == api_key
    assert processor.model.model_name == "models/gemini-1.5-flash-latest"

    response = processor.generate_response("Hello, how are you?")
    assert isinstance(response, str)


# ClaudePromptProcessor.py

from typing import List, Union, Optional
from AbstractPromptProcessor import AbstractPromptProcessor

try:
    import anthropic
except ImportError:
    anthropic = None

class ClaudePromptProcessor(AbstractPromptProcessor):
    def __init__(self, api_key: Optional[str] = None, model: str = "claude-3-haiku-20240307"):
        super().__init__(api_key)
        self.model = model
        
    def generate_response(
        self,
        prompt: Union[str, List[str]],
        *,
        model: Optional[str] = "claude-3-haiku-20240307",
        max_tokens: Optional[int] = 4096,
        temperature: float = 1.0,
        top_p: float = 1.0,
        logprobs: Optional[int] = None,
        stream: bool = False,
        system: Optional[str] = None,
        # Claude specific arguments
        metadata: Optional[dict] = None,
        stop_sequences: Optional[List[str]] = None,
        tool_choice: Optional[Union[str, dict]] = None,
        tools: Optional[List[dict]] = None,
        top_k: Optional[int] = None,
        user: Optional[str] = None,
    ) -> Union[str, List[str]]:
        if anthropic is None:
            raise ImportError("The 'anthropic' library is not installed. Please install it to use ClaudePromptProcessor.")
        if model is None:
            model = self.model
        
        kwargs = {
            "model": model,
            "max_tokens": max_tokens,
            "temperature": temperature,
            "top_p": top_p,
            "stop_sequences": stop_sequences,
            "logprobs": logprobs,
            "metadata": metadata,
            "system": super().remove_special_characters(system),
            "tool_choice": tool_choice,
            "tools": tools,
            "top_k": top_k,
            "user_id": user,
        }
        
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        
        messages = []
        if isinstance(prompt, str):
            messages.append({"role": "user", "content": super().remove_special_characters(prompt)})
        else:
            messages.extend([{"role": "user", "content": super().remove_special_characters(p)} for p in prompt])

        response = anthropic.Anthropic().messages.create(messages=messages, **kwargs) if stream else anthropic.Anthropic().messages.create(messages=messages, **kwargs)
        
        return response.content[0].text


# GeminiPromptProcessor.py

from typing import List, Union, Optional
from AbstractPromptProcessor import AbstractPromptProcessor

try:
    import google.generativeai as genai
    import google.generativeai.types as genai_types
    from google.generativeai.types import HarmCategory, HarmBlockThreshold
    import google.ai.generativelanguage as genai_lang
except ImportError:
    genai = None

class GeminiPromptProcessor(AbstractPromptProcessor):
    def __init__(self, api_key: Optional[str] = None, model: str = "models/gemini-1.5-flash-latest"):
        super().__init__(api_key)
        genai.configure(api_key=api_key)
        self.model = genai.GenerativeModel(model)

    def generate_response(
        self,
        prompt: Union[str, List[str]],
        *,
        model: Optional[str] = "models/gemini-1.5-flash-latest",
        max_tokens: Optional[int] = None,
        temperature: float = 1.0,
        top_p: float = 1.0,
        top_k: int = 1,
        n: int = 1,
        stop: Optional[Union[str, List[str]]] = None,
        stream: bool = False,
        system: Optional[str] = None,
        # Gemini specific arguments
        safety_settings = {
            HarmCategory.HARM_CATEGORY_HATE_SPEECH: HarmBlockThreshold.BLOCK_NONE,
            HarmCategory.HARM_CATEGORY_HARASSMENT: HarmBlockThreshold.BLOCK_NONE,
            HarmCategory.HARM_CATEGORY_SEXUALLY_EXPLICIT: HarmBlockThreshold.BLOCK_NONE,
            HarmCategory.HARM_CATEGORY_DANGEROUS_CONTENT: HarmBlockThreshold.BLOCK_NONE,
            # Not supported in Gemini models
            # HarmCategory.HARM_CATEGORY_UNSPECIFIED: HarmBlockThreshold.BLOCK_NONE,
            # HarmCategory.HARM_CATEGORY_DEROGATORY: HarmBlockThreshold.BLOCK_NONE,
            # HarmCategory.HARM_CATEGORY_TOXICITY: HarmBlockThreshold.BLOCK_NONE,
            # HarmCategory.HARM_CATEGORY_VIOLENCE: HarmBlockThreshold.BLOCK_NONE,
            # HarmCategory.HARM_CATEGORY_SEXUAL: HarmBlockThreshold.BLOCK_NONE,
            # HarmCategory.HARM_CATEGORY_MEDICAL: HarmBlockThreshold.BLOCK_NONE,
            # HarmCategory.HARM_CATEGORY_DANGEROUS: HarmBlockThreshold.BLOCK_NONE,
        },
        generation_config: Optional[genai.GenerationConfig] = None,
        tools: Optional[genai_types.FunctionLibrary] = None,
        tool_config: Optional[genai_lang.ToolConfig] = None,
    ) -> Union[str, List[str]]:
        if genai is None:
            raise ImportError("The 'google.generativeai' library is not installed. Please install it to use GeminiPromptProcessor.")
        
        if model is not None or system is not None:
            self.model = genai.GenerativeModel(model, system_instruction=super().remove_special_characters(system))

        contents = []
        if isinstance(prompt, str):
            contents.append(super().remove_special_characters(prompt))
        else:
            contents.extend([super().remove_special_characters(p) for p in prompt])

        if generation_config is None:
            generation_config = genai.GenerationConfig(
                    candidate_count=n,
                    stop_sequences=stop,
                    max_output_tokens=max_tokens,
                    temperature=temperature,
                    top_p=top_p,
                    top_k=top_k,
            )

        response = self.model.generate_content(
            contents=contents,
            stream=stream,
            safety_settings=safety_settings,
            generation_config=generation_config,
            tools=tools,
            tool_config=tool_config,
        )

        return response.text


from abc import ABC, abstractmethod
from typing import List, Union, Optional
from dotenv import load_dotenv
import os

class AbstractPromptProcessor(ABC):
    def __init__(self, api_key: Optional[str] = None, model: Optional[str] = None):
        self._api_key = api_key

    @abstractmethod
    def generate_response(
        self,
        prompt: Union[str, List[str]],
        *,
        model: str,
        max_tokens: Optional[int] = None,
        temperature: float = 1.0,
        top_p: float = 1.0,
        n: int = 1,
        stop: Optional[Union[str, List[str]]] = None,
        presence_penalty: float = 0.0,
        frequency_penalty: float = 0.0,
        logprobs: Optional[int] = None,
        stream: bool = False,
        system: Optional[str] = None,
    ) -> Union[str, List[str]]:
        """
        Generate a response from the language model.

        Args:
            prompt: The input prompt or a list of prompts.
            model: The name of the model to use.
            max_tokens: The maximum number of tokens to generate.
            temperature: The sampling temperature for controlling randomness.
            top_p: The cumulative probability threshold for nucleus sampling.
            n: The number of responses to generate.
            stop: The stopping sequence(s) for generation.
            presence_penalty: The penalty for repeated tokens.
            frequency_penalty: The penalty for frequent tokens.
            logprobs: The number of logprobs to return.
            stream: Whether to stream the response.
            system: System Prompt.

        Returns:
            The generated response or a list of responses.
        """
        pass

    @property
    def api_key(self) -> str:
        """Get the API key."""
        return self._api_key

    @api_key.setter
    def api_key(self, value: str):
        """Set the API key."""
        self._api_key = value

    def load_api_key(self, env_var: str):
        """
        Load the API key from an environment variable using dotenv.

        Args:
            env_var: The name of the environment variable containing the API key.
        """
        load_dotenv()
        self._api_key = os.getenv(env_var)
        if self._api_key is None:
            raise ValueError(f"API key not found in environment variable {env_var}")
        
    def remove_special_characters(self, text):
        if text is None:
            return text
        text = text.replace("<|", "")
        text = text.replace("|>", "")
        return text


from setuptools import setup, find_packages

setup(
    name='FLUTE-LLM',
    version='0.1.0',
    description='A package for prompt processing and language model interaction',
    author='The Pioneer',
    url='https://github.com/thepioneerjp/FLUTE',
    packages=find_packages(),
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    install_requires=[
        'openai',
        'anthropic',
        'google-generativeai',
        'python-dotenv',
        'pytest'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
    ],
)


# GPTPromptProcessor.py

from typing import List, Union, Optional
from AbstractPromptProcessor import AbstractPromptProcessor

try:
    import openai
except ImportError:
    openai = None

class GPTPromptProcessor(AbstractPromptProcessor):
    def __init__(self, api_key: Optional[str] = None, organization: Optional[str] = None, model: str = "gpt-4o"):
        super().__init__(api_key)
        self.organization = organization

    def generate_response(
        self,
        prompt: Union[str, List[str]],
        *,
        model: str = "gpt-4o",
        max_tokens: Optional[int] = 4096,
        temperature: float = 1.0,
        top_p: float = 1.0,
        n: int = 1,
        stop: Optional[Union[str, List[str]]] = None,
        logprobs: Optional[int] = None,
        stream: bool = False,
        system: Optional[str] = None,
        # GPT specific arguments
        presence_penalty: float = 0.0,
        frequency_penalty: float = 0.0,
        logit_bias: Optional[dict] = None,
        top_logprobs: Optional[int] = None,
        response_format: Optional[dict] = None,
        seed: Optional[int] = None,
        stream_options: Optional[dict] = None,
        tools: Optional[List[dict]] = None,
        tool_choice: Optional[Union[str, dict]] = None,
        user: Optional[str] = None,
    ) -> Union[str, List[str]]:
        if openai is None:
            raise ImportError("The 'openai' library is not installed. Please install it to use GPTPromptProcessor.")
        openai.api_key = self.api_key
        if self.organization:
            openai.organization = self.organization

        messages = []
        if system is not None:
            messages.append({"role": "system", "content": super().remove_special_characters(system)})
        if isinstance(prompt, str):
            messages.append({"role": "user", "content": super().remove_special_characters(prompt)})
        else:
            messages.extend([{"role": "user", "content": super().remove_special_characters(p)} for p in prompt])

        kwargs = {
            "model": model,
            "messages": messages,
            "max_tokens": max_tokens,
            "temperature": temperature,
            "top_p": top_p,
            "n": n,
            "stop": stop,
            "presence_penalty": presence_penalty,
            "frequency_penalty": frequency_penalty,
            "logprobs": logprobs,
            "stream": stream,
            "logit_bias": logit_bias,
            "top_logprobs": top_logprobs,
            "response_format": response_format,
            "seed": seed,
            "stream_options": stream_options,
            "tools": tools,
            "tool_choice": tool_choice,
            "user": user,
        }

        kwargs = {k: v for k, v in kwargs.items() if v is not None}

        response = openai.chat.completions.create(**kwargs)

        return response.choices[0].message.content.strip()


# PromptProcessorFactory.py

from typing import Optional

from AbstractPromptProcessor import AbstractPromptProcessor
from ClaudePromptProcessor import ClaudePromptProcessor
from GPTPromptProcessor import GPTPromptProcessor
from GeminiPromptProcessor import GeminiPromptProcessor

class PromptProcessorFactory:
    @staticmethod
    def create_prompt_processor(model_name: str, api_key: Optional[str] = None):
        if model_name in [
            "claude-3-opus-20240229",
            "claude-3-sonnet-20240229",
            "claude-3-haiku-20240307",
        ]:
            return ClaudePromptProcessor(api_key=api_key, model=model_name)
        elif model_name in [
            "gpt-4o",
            "gpt-4o-2024-05-13",
            "gpt-4-turbo",
            "gpt-4-turbo-2024-04-09",
            "gpt-4-turbo-preview", 
            "gpt-4-0125-preview",
            "gpt-4-1106-preview",
            "gpt-4-vision-preview",
            "gpt-4-1106-vision-preview",
            "gpt-4",
            "gpt-4-0613",
            "gpt-4-32k",
            "gpt-4-32k-0613",
        ]:
            return GPTPromptProcessor(api_key=api_key)
        elif model_name in [
            "models/gemini-1.5-pro-latest", 
            "models/gemini-1.5-flash-latest",
            "models/gemini-pro",
            "models/gemini-pro-vision",
        ]:
            return GeminiPromptProcessor(api_key=api_key, model=model_name)
        ## TODO: Groq, Local LLMs, Brain Computer Interface, etc...
        else:
            return None